import pathlib
import numpy as np
import pandas as pd

from pyometiff import OMETIFFReader, OMETIFFWriter
from scipy.ndimage import gaussian_filter
from skimage.filters import threshold_otsu
from skimage.measure import label, regionprops
from skimage.segmentation import find_boundaries
from skimage.morphology import ball, opening, disk, ellipse
from scipy.spatial import cKDTree
from concurrent.futures import ThreadPoolExecutor


def generate_and_visualize_future(filePath : pathlib.Path):
    future = ThreadPoolExecutor().submit(generate_and_visualize,filePath)
    return future


def generate_and_visualize(filePath : pathlib.Path):
    load_and_generate_data(filePath)
    return get_layers_to_visualize(filePath)


def load_and_generate_data(filePath : pathlib.Path):

    # Load the full z stack, this will be handled later by napari
    reader = OMETIFFReader(fpath=filePath)
    img_array, metadata, xml_metadata = reader.read()

    channel_0 = img_array[0]  # shape: (140, 512, 512)
    channel_1 = img_array[1]  # shape: (140, 512, 512)

    ch1_filtered = gaussian_filter(channel_1, sigma=[2.,1.,1.])
    filt_op = np.max(ch1_filtered, axis=2)
    thresh = threshold_otsu(filt_op)

    # Apply threshold and label
    binary = ch1_filtered > thresh
    blobs = label(binary)
    assert blobs.max()<65000, "issues here"
    blobs = blobs.astype(np.uint16)

    # `blobs` is a labeled 3D array
    props = regionprops(blobs)

    # Extract centroids
    center_coords = np.array([p.centroid for p in props])  # shape: (N, 3), in (z, y, x)

    fname = filePath.name
    fdir = filePath.absolute()
    fname = fname.split(".ome.tiff")[0]
    fname = fname+"_blobs.ome.tiff"
    omePath = filePath.parent.joinpath(fname)
    print(omePath)

    dimension_order = "ZYX"
    metadata_dict = {
        "PhysicalSizeX" : metadata['PhysicalSizeX'],
        "PhysicalSizeXUnit" : "µm",
        "PhysicalSizeY" : metadata['PhysicalSizeY'],
        "PhysicalSizeXUnit" : "µm",
        "PhysicalSizeZ" : metadata['PhysicalSizeZ'],
        "PhysicalSizeZUnit" : "µm",
        
        "Channels" : {
            "filtered" : {
                "Name" : "blobs",
                "SamplesPerPixel": 1,
            },
        }
    }

    writer = OMETIFFWriter(
        fpath=omePath,
        dimension_order=dimension_order,
        array=blobs,
        metadata=metadata_dict,
        explicit_tiffdata=False)

    writer.write()


    def keep_largest_label(label_img):
        props = regionprops(label_img)
        if not props:
            return np.zeros_like(label_img, dtype=label_img.dtype)

        # Find the region with the largest area
        largest_region = max(props, key=lambda r: r.area)
        largest_label = largest_region.label

        # Create a mask with only the largest label
        return (label_img == largest_label).astype(label_img.dtype)


    ch0_filtered = gaussian_filter(channel_0, sigma=[.3,2,2])
    filt_op = np.mean(ch0_filtered, axis=2)
    thresh = threshold_otsu(filt_op)

    # Apply threshold and label
    surface_bw = ch0_filtered > thresh

    selem = ellipse(5,1)
    img = surface_bw[:,:,0]
    opened = opening(img, selem)

    opened_volume = np.empty_like(surface_bw)

    for x in range(surface_bw.shape[2]):
        img = surface_bw[:,:,x]
        opened = opening(img, selem)
        opened_volume[:,:,x] = opened

    for y in range(opened_volume.shape[2]):
        img = opened_volume[:,y,:]
        opened = opening(img, selem)
        opened_volume[:,y,:] = opened

    surface = label(opened_volume)
    assert surface.max()<65000, "issues here"
    surface = surface.astype(np.uint16)

    surface = keep_largest_label(surface)
    fname = filePath.name
    fname = fname.split(".ome.tiff")[0]
    fname = fname+"_surface.ome.tiff"
    omePath = filePath.parent.joinpath(fname)
    print(omePath)

    dimension_order = "ZYX"
    metadata_dict = {
        "PhysicalSizeX" : metadata['PhysicalSizeX'],
        "PhysicalSizeXUnit" : "µm",
        "PhysicalSizeY" : metadata['PhysicalSizeY'],
        "PhysicalSizeXUnit" : "µm",
        "PhysicalSizeZ" : metadata['PhysicalSizeZ'],
        "PhysicalSizeZUnit" : "µm",
        
        "Channels" : {
            "filtered" : {
                "Name" : "surface",
                "SamplesPerPixel": 1,
            },
        }
    }

    writer = OMETIFFWriter(
        fpath=omePath,
        dimension_order=dimension_order,
        array=surface,
        metadata=metadata_dict,
        explicit_tiffdata=False)

    writer.write()


    # Physical voxel size: (z, y, x)
    voxel_size = np.array([metadata['PhysicalSizeZ'], metadata['PhysicalSizeY'], metadata['PhysicalSizeX']])  # in microns

    edges = find_boundaries(surface>0, mode='outer')
    edge_coords = np.column_stack(np.nonzero(edges))

    # Scale coordinates to real physical units
    edge_coords_phys = edge_coords * voxel_size
    center_coords_phys = center_coords * voxel_size

    # Build KD-tree in physical space
    tree = cKDTree(edge_coords_phys)
    dists, indices = tree.query(center_coords_phys, k=1)

    # Find closest edge point for each center
    nearest_edge_coords = edge_coords[indices]  # shape: (n, 3)
    nearest_edge_coords_phys = edge_coords_phys[indices]

    # Compute displacement vectors
    displacements = nearest_edge_coords - center_coords  # shape: (n, 3)
    displacements_phys = nearest_edge_coords_phys - center_coords_phys  # shape: (n, 3)

    # Build table
    vector_table = pd.DataFrame({
        'z0': center_coords[:, 0],
        'y0': center_coords[:, 1],
        'x0': center_coords[:, 2],
        'z1': nearest_edge_coords[:, 0],
        'y1': nearest_edge_coords[:, 1],
        'x1': nearest_edge_coords[:, 2],
        'dz': displacements[:, 0],
        'dy': displacements[:, 1],
        'dx': displacements[:, 2],
        'dz_phys': displacements_phys[:, 0],
        'dy_phys': displacements_phys[:, 1],
        'dx_phys': displacements_phys[:, 2],

        'distance': dists
    })

    # Save to CSV
    fname = filePath.name
    fname = fname.split(".ome.tiff")[0]
    fname = fname + "_vector_table.csv"

    csv_path = filePath.parent.joinpath(fname)
    vector_table.to_csv(csv_path, index=False)
    
    
def get_layers_to_visualize(imgPath : pathlib.Path):
    # Load the full z stack, this will be handled later by napari
    img_name = imgPath.name
    img_name = img_name.split(".ome.tiff")[0]

    reader = OMETIFFReader(fpath=imgPath)
    img_array, metadata, xml_metadata = reader.read()
    nframes = img_array.shape[0]

    channel_0 = img_array[0]  # shape: (140, 512, 512)
    channel_1 = img_array[1]  # shape: (140, 512, 512)

    channel_0.shape

    surface_path = pathlib.Path(f"./data/{img_name}_surface.ome.tiff")
    reader = OMETIFFReader(fpath=surface_path)
    surface, metadata, xml_metadata = reader.read()
    surface.shape

    blobs_path = pathlib.Path(f"./data/{img_name}_blobs.ome.tiff")
    reader = OMETIFFReader(fpath=blobs_path)
    blobs, metadata, xml_metadata = reader.read()

    df = pd.read_csv(f"./data/{img_name}_vector_table.csv")

    # Reconstruct Napari vectors array
    starts = df[['z0', 'y0', 'x0']].to_numpy()
    ends   = df[['z1', 'y1', 'x1']].to_numpy()
    ends = np.ones_like(starts)*30
    shifts = df[['dz', 'dy', 'dx']].to_numpy()
    vectors = np.stack([starts.astype(np.float32), shifts.astype(np.float32)], axis=1)  # shape: (n, 2, 3)

    voxel_size = np.array([metadata['PhysicalSizeZ'], metadata['PhysicalSizeY'], metadata['PhysicalSizeX']])  # in microns

    return channel_0, "image", surface, "surface", blobs, "blobs", vectors,"vectors", voxel_size



from typing import TYPE_CHECKING

from qtpy.QtCore import Signal, Qt
from qtpy.QtWidgets import QVBoxLayout, QPushButton, QWidget, QFileDialog, QProgressDialog
from .generate_data import load_and_generate_data, generate_and_visualize_future
import numpy as np
from pathlib import Path

if TYPE_CHECKING:
    import napari

class MainWidget(QWidget):
    
    
    show_all_layers_signal = Signal( np.ndarray, str, np.ndarray, str, np.ndarray, str, np.ndarray, str, int)
    close_progress_signal = Signal()
    # use a type annotation of 'napari.viewer.Viewer' for any parameter
    def __init__(self, viewer: "napari.viewer.Viewer"):
        super().__init__()
        self.viewer = viewer

        readDataBtn = QPushButton("Open file to visualize...")
        readDataBtn.clicked.connect(self._on_open_file)

        self.progress_dialog = QProgressDialog("","",0,0)
        self.progress_dialog.setAutoClose(True)
        self.progress_dialog.reset()
        
        self.setLayout(QVBoxLayout())
        self.layout().addWidget(readDataBtn)
        self.dataFile = None
        
        self.show_all_layers_signal.connect(self._show_all_layers,Qt.QueuedConnection)
        self.close_progress_signal.connect(self._close_progress_dialog, Qt.QueuedConnection)

    def _on_open_file(self):
        data = QFileDialog.getOpenFileName()
        if not data[0]:
            return
        self.dataFile = data[0]
        
        self.run_all(self.dataFile)
        
    def _show_all_layers(self, i,ii , s,si , b,bi , v,vi , scale):
        
        self.viewer.add_image(i)
        self.viewer.add_labels(s, name=si)
        self.viewer.add_labels(b, name=bi)
        self.viewer.add_vectors(v, name=vi, edge_color='lime')
        
        #set 3d view mode
        self.viewer.dims.ndisplay = 3
        
    def run_all(self, file: Path):
        future = generate_and_visualize_future(Path(file))
        
        f_cb = self._future_done_callback
        future.add_done_callback(f_cb)
        self.show_progress_dialog("Doing calculation ...","cancel")


    def _future_done_callback(self, future, success_msg = None, error_msg = None, show_exception = True):
        self.close_progress_signal.emit()

        msg = success_msg
        title = "Success"
        # The task is completed
        if future.done():
            if future.exception() is not None and error_msg is not None:
                msg = error_msg
                title = "Error during task"
                if show_exception:
                    msg += ": " + str(future.exception())
                
            else:
                self.show_all_layers_signal.emit(*future.result())
                self.close_progress_signal.emit()
 
    def _close_progress_dialog(self):
        self.progress_dialog.reset()

    def show_progress_dialog(self,label,cancel):
        self.progress_dialog.setLabelText(label)
        self.progress_dialog.setCancelButtonText(cancel)
        self.progress_dialog.open()




# def get_layers_to_visualize(imgPath : Path):
#     # Load the full z stack, this will be handled later by napari
#     #img_fpath = Path("./data/flat-01.ome.tiff")
#     #img_fpath = Path("./data/193_S_CTRL_Before.ome.tiff")
#     img_name = imgPath.name
#     img_name = img_name.split(".ome.tiff")[0]

#     #img_fpath = Path("./data/curved-01.ome.tiff")
#     reader = OMETIFFReader(fpath=imgPath)
#     img_array, metadata, xml_metadata = reader.read()
#     nframes = img_array.shape[0]
#     # img_array.shape

#     channel_0 = img_array[0]  # shape: (140, 512, 512)
#     channel_1 = img_array[1]  # shape: (140, 512, 512)

#     channel_0.shape

#     surface_path = Path(f"./data/{img_name}_surface.ome.tiff")
#     reader = OMETIFFReader(fpath=surface_path)
#     surface, metadata, xml_metadata = reader.read()
#     surface.shape

#     blobs_path = Path(f"./data/{img_name}_blobs.ome.tiff")
#     reader = OMETIFFReader(fpath=blobs_path)
#     blobs, metadata, xml_metadata = reader.read()

#     df = pd.read_csv(f"./data/{img_name}_vector_table.csv")

#     # Reconstruct Napari vectors array
#     starts = df[['z0', 'y0', 'x0']].to_numpy()
#     ends   = df[['z1', 'y1', 'x1']].to_numpy()
#     ends = np.ones_like(starts)*30
#     shifts = df[['dz', 'dy', 'dx']].to_numpy()
#     vectors = np.stack([starts.astype(np.float32), shifts.astype(np.float32)], axis=1)  # shape: (n, 2, 3)

#     voxel_size = np.array([metadata['PhysicalSizeZ'], metadata['PhysicalSizeY'], metadata['PhysicalSizeX']])  # in microns
#     #scaling = voxel_size/voxel_size.max()
#     # for layer in viewer.layers:
#     #     layer.scale = voxel_size

#     return (channel_0, "image"), (surface, "surface"), (blobs, "blobs"), (vectors,"vectors"), voxel_size

#     # vectors.shape

#     # Start Napari viewer
#     # viewer = napari.Viewer()

#     # # Add the image to the viewer
#     # viewer.add_image(channel_0)
#     # viewer.add_labels(surface, name="surface")
#     # viewer.add_labels(blobs, name="blobs")
#     # viewer.add_vectors(vectors, name="vectors", edge_color='lime')





try:
    from ._version import version as __version__
except ImportError:
    __version__ = "0.0.1"

# from ._reader import napari_get_reader
# from ._sample_data import make_sample_data
from ._widget import MainWidget
# from ._writer import write_multiple, write_single_image

__all__ = (
#    "napari_get_reader",
    # "write_single_image",
    # "write_multiple",
    # "make_sample_data",
    "MainWidget",
)



# launch_napari.py
import napari
#from napari import Viewer, run

viewer = napari.Viewer()
dock_widget, plugin_widget = viewer.window.add_plugin_dock_widget(
    "brendan-beads", ""
)
# Optional steps to setup your plugin to a state of failure
# E.g. plugin_widget.parameter_name.value = "some value"
# E.g. plugin_widget.button.click()
napari.run()